"""utility methods to do logging. this is more of convenience and very likely to be replaced with better implementation"""

import urllib
import urllib.parse
import json
from json.decoder import JSONDecodeError
import os
import math
import mmh3
import time
from concurrent.futures import ThreadPoolExecutor

# TODO: these caches dont work in multithreaded env.
MSG_CACHE_MAX_LEN = 10000
INFO_MSG_CACHE = {}
ERROR_MSG_CACHE = {}
WARN_MSG_CACHE = {}
DEBUG_MSG_CACHE = {}
TRACE_MSG_CACHE = {}

# some env variables
OMIGO_CRITICAL = "OMIGO_CRITICAL"
OMIGO_ERROR = "OMIGO_ERROR"
OMIGO_WARN = "OMIGO_WARN"
OMIGO_INFO = "OMIGO_INFO"
OMIGO_DEBUG = "OMIGO_DEBUG"
OMIGO_TRACE = "OMIGO_TRACE"
OMIGO_WARN_ONCE_ONLY = "OMIGO_WARN_ONCE_ONLY"
OMIGO_DEBUG_REPORT_PROGRESS_PERC = "OMIGO_DEBUG_REPORT_PROGRESS_PERC"
OMIGO_DEBUG_REPORT_PROGRESS_MIN_THRESH = "OMIGO_DEBUG_REPORT_PROGRESS_MIN_THRESH"
OMIGO_CODE_TODO_WARNING = "OMIGO_CODE_TODO_WARNING"
OMIGO_BIG_TSV_WARN_SIZE_THRESH = "OMIGO_BIG_TSV_WARN_SIZE_THRESH"

def is_critical():
    return str(os.environ.get(OMIGO_CRITICAL, "1")) == "1"

def is_error():
    return str(os.environ.get(OMIGO_ERROR, "1")) == "1"

def is_warn():
    return str(os.environ.get(OMIGO_WARN, "1")) == "1"

def is_info():
    return str(os.environ.get(OMIGO_INFO, "1")) == "1"

def is_debug():
    return str(os.environ.get(OMIGO_DEBUG, "0")) == "1"

def is_trace():
    return str(os.environ.get(OMIGO_TRACE, "0")) == "1"

def is_warn_once_only():
    return str(os.environ.get(OMIGO_WARN_ONCE_ONLY, "0")) == "1"

def get_report_progress():
    return float(os.environ.get(OMIGO_DEBUG_REPORT_PROGRESS_PERC, "0"))

def get_report_progress_min_thresh():
    return float(os.environ.get(OMIGO_DEBUG_REPORT_PROGRESS_MIN_THRESH, "100000"))

def set_report_progress_perc(perc):
    os.environ[OMIGO_DEBUG_REPORT_PROGRESS_PERC] = str(perc)

def set_report_progress_min_thresh(thresh):
    os.environ[OMIGO_DEBUG_REPORT_PROGRESS_MIN_THRESH] = str(thresh)

def get_big_tsv_warn_size_thresh():
    return float(os.environ.get(OMIGO_BIG_TSV_WARN_SIZE_THRESH, "1000000000"))

def set_big_tsv_warn_size_thresh(thresh):
    os.environ[OMIGO_BIG_TSV_WARN_SIZE_THRESH] = str(thresh)

def trace(msg):
    if (is_trace()):
        print("[TRACE]: {}".format(msg))

def trace_once(msg):
    # check if enabled
    if (is_trace() == False):
        return

    # refer to global variable
    global TRACE_MSG_CACHE
    # check if msg is already displayed
    if (msg not in TRACE_MSG_CACHE.keys()):
        print("[TRACE ONCE ONLY]: " + msg)
        TRACE_MSG_CACHE[msg] = 1

        # check if the cache has become too big
        if (len(TRACE_MSG_CACHE) >= MSG_CACHE_MAX_LEN):
            TRACE_MSG_CACHE = {}

def debug(msg):
    if (is_debug()):
        print("[DEBUG]: {}".format(msg))

def debug_once(msg):
    # check if enabled
    if (is_debug() == False):
        return

    # refer to global variable
    global DEBUG_MSG_CACHE
    # check if msg is already displayed
    if (msg not in DEBUG_MSG_CACHE.keys()):
        print("[DEBUG ONCE ONLY]: " + msg)
        DEBUG_MSG_CACHE[msg] = 1

        # check if the cache has become too big
        if (len(DEBUG_MSG_CACHE) >= MSG_CACHE_MAX_LEN):
            DEBUG_MSG_CACHE = {}

def info(msg):
    if (is_info()):
        print("[INFO]: {}".format(msg))

def info_once(msg):
    # check if enabled
    if (is_info() == False):
        return

    # refer to global variable
    global INFO_MSG_CACHE
    # check if msg is already displayed
    if (msg not in INFO_MSG_CACHE.keys()):
        print("[INFO ONCE ONLY]: " + msg)
        INFO_MSG_CACHE[msg] = 1

        # check if the cache has become too big
        if (len(INFO_MSG_CACHE) >= MSG_CACHE_MAX_LEN):
            INFO_MSG_CACHE = {}

def error(msg):
    if (is_error()):
        print("[ERROR]: {}".format(msg))

def error_once(msg):
    # check if enabled
    if (is_error() == False):
        return

    # check if msg is already displayed
    if (msg not in ERROR_MSG_CACHE.keys()):
         print("[ERROR ONCE ONLY]: {}".format(msg))
         ERROR_MSG_CACHE[msg] = 1

         # clear the cache if it has become too big
         if (len(ERROR_MSG_CACHE) >= MSG_CACHE_MAX_LEN):
             ERROR_MSG_CACHE = {}
    else:
        trace(msg)

def enable_critical_mode():
    os.environ[OMIGO_CRITICAL] = "1"

def enable_error_mode():
    os.environ[OMIGO_ERROR] = "1"

def enable_warn_mode():
    os.environ[OMIGO_WARN] = "1"

def enable_info_mode():
    os.environ[OMIGO_INFO] = "1"

def enable_debug_mode():
    os.environ[OMIGO_DEBUG] = "1"

def enable_trace_mode():
    os.environ[OMIGO_TRACE] = "1"

def disable_critical_mode():
    os.environ[OMIGO_CRITICAL] = "0"

def disable_error_mode():
    os.environ[OMIGO_ERROR] = "0"

def disable_warn_mode():
    os.environ[OMIGO_WARN] = "0"

def disable_info_mode():
    os.environ[OMIGO_INFO] = "0"

def disable_debug_mode():
    os.environ[OMIGO_DEBUG] = "0"

def disable_trace_mode():
    os.environ[OMIGO_TRACE] = "0"

def enable_warn_once_only_mode():
    os.environ[OMIGO_WARN_ONCE_ONLY] = "1"

def disable_warn_once_only_mode():
    os.environ[OMIGO_WARN_ONCE_ONLY] = "1"

def warn(msg):
    if (is_warn() or is_error() or is_critical()):
        print("[WARN]: " + msg)

def warn_once(msg):
    # check if enabled
    if (is_warn_once_only() == False):
        return

    # refer to global variable
    global WARN_MSG_CACHE
    # check if msg is already displayed
    if (msg not in WARN_MSG_CACHE.keys()):
        print("[WARN ONCE ONLY]: " + msg)
        WARN_MSG_CACHE[msg] = 1

        # check if the cache has become too big
        if (len(WARN_MSG_CACHE) >= MSG_CACHE_MAX_LEN):
            WARN_MSG_CACHE = {}

def is_code_todo_warning():
    return str(os.environ.get(OMIGO_CODE_TODO_WARNING, "0")) == "1"

def print_code_todo_warning(msg):
    if (is_code_todo_warning()):
        print("[CODE TODO WARNING]: " + msg)

def url_encode(s):
    if (s is None):
        return ""

    return urllib.parse.quote_plus(s)

# TODO: this replaces TAB character
def url_decode(s):
    if (s is None):
        return ""

    return urllib.parse.unquote_plus(s).replace("\n", " ").replace("\t", " ").replace("\v", " ").replace("\r", " ")

def url_decode_clean(s):
    return url_decode(s).replace("\n", " ").replace("\v", " ").replace("\r", " ").replace("\t", " ")

# move this to utils
def parse_encoded_json(s):
    if (s is None):
        return {}

    if (s == ""):
        return {}

    try:
        decoded = url_decode(s)
        if (decoded == ""):
            return {}

        return json.loads(decoded)
    except JSONDecodeError:
        print("Error in decoding json")
        return {}

def encode_json_obj(json_obj):
    return url_encode(json.dumps(json_obj))

# TODO: make it more robust. The object key doesnt have / prefix or suffix
def split_s3_path(path):
    part1 = path[5:]
    index = part1.index("/")
    bucket_name = part1[0:index]

    # boundary conditions
    if (index < len(part1) - 1):
        object_key = part1[index+1:]
    else:
        object_key = ""

    # remove trailing suffix
    if(object_key.endswith("/")):
        object_key = object_key[0:-1]

    return bucket_name, object_key

def get_counts_map(xs):
    mp = {}
    for x in xs:
        if (x not in mp):
            mp[x] = 0
        mp[x] = mp[x] + 1

    return mp

def report_progress(msg, dmsg, counter, total):
    report_progress = get_report_progress()
    report_progress_min_threshold = get_report_progress_min_thresh()
    msg2 = dmsg + ": " + msg if (dmsg is not None and len(dmsg) > 0) else msg
    if (is_debug() and report_progress > 0 and total >= report_progress_min_threshold):
        progress_size = int(report_progress * total)
        if (progress_size > 0 and counter % progress_size == 0):
            progress_perc = int(math.ceil(100 * counter / total))
            debug("{}: {}% ({} / {})".format(msg2, progress_perc, counter, total))

def merge_arrays(arr_list):
    result = []
    for arr in arr_list:
        for v in arr:
            result.append(v)

    return result

def is_array_of_string_values(col_or_cols):
    if (isinstance(col_or_cols, str)):
        return False

    is_array = False
    for c in col_or_cols:
        if (len(c) > 1):
            is_array = True
            break
    return is_array

def is_numeric(v):
    return str(v).isnumeric()

def is_float(v):
    try:
        float(str(v))
        return True
    except:
        return False

def is_int_col(xtsv, col):
    raise Exception("Deprecated. The to_df() method supports inferring data types")

def is_float_col(xtsv, col):
    raise Exception("Deprecated. The to_df() method supports inferring data types")

def is_pure_float_col(xtsv, col):
    raise Exception("Deprecated. The to_df() method supports inferring data types")

def is_float_with_fraction(xtsv, col):
    raise Exception("Deprecated. The to_df() method supports inferring data types")

def compute_hash(x, seed = 0):
    return abs(mmh3.hash64(str(x) + str(seed))[0])

class ThreadPoolTask:
    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs

def run_with_thread_pool(tasks, num_par = 4, wait_sec = 10, post_wait_sec = 0, dmsg = ""):
    dmsg = extend_inherit_message(dmsg, "run_with_thread_pool")

    # debug
    if (num_par > 0):
        info("{}: num tasks: {}, num_par: {}".format(dmsg, len(tasks), num_par))

    # define results
    results = []

    # check if this is to be run in multi threaded mode or not
    if (num_par == 0):
        debug("{}: running in single threaded mode".format(dmsg))

        # iterate
        for task in tasks:
            func = task.func
            args = task.args
            kwargs = task.kwargs
            results.append(func(*args, **kwargs))

        # return
        return results
    else:
        # start thread pool
        future_results = []
        with ThreadPoolExecutor(max_workers = num_par) as executor:
            # iterate over tasks and call submit
            for i in range(len(tasks)):
                # call submit
                func = tasks[i].func
                args = tasks[i].args
                kwargs = tasks[i].kwargs
                future_results.append(executor.submit(func, *args, **kwargs))

            # wait for completion
            while True:
                done_count = 0
                for f in future_results:
                    if (f.done() == True):
                        done_count = done_count + 1

                # check if all are done
                if (done_count < len(future_results)):
                    # sleep for some additional time to allow notebook stop method to work
                    info("{}: futures not completed yet. Status: {} / {}. Sleeping for {} sec".format(dmsg, done_count, len(future_results), wait_sec))
                    time.sleep(wait_sec)
                else:
                    info("{}: run_with_thread_pool: finished".format(dmsg))
                    break

            # combine the results
            for f in future_results:
                results.append(f.result())

            # wait for post_wait_sec for mitigating eventual consistency
            if (post_wait_sec > 0):
                info("{}: sleeping for post_wait_sec: {}".format(dmsg, post_wait_sec))
                time.sleep(post_wait_sec)

            # return
            return results

def raise_exception_or_warn(msg, ignore_if_missing, max_len = 2000):
    # strip the message to max_len
    if (max_len is not None and max_len > 0 and len(msg) > max_len):
        msg = msg[0:max_len] + " ..."

    # print message if ignore_if_missing flag is true
    if (ignore_if_missing == True):
        debug_once(msg)
    else:
        raise Exception(msg)

def error_and_raise_exception(msg, max_len = 2000):
    # strip the message to max_len
    if (max_len is not None and max_len > 0 and len(msg) > max_len):
        msg = msg[0:max_len] + " ..."

    # print error
    error(msg)

    # raise exception
    raise Exception(msg)

# Deprecated
def strip_spl_white_spaces(v):
    warn_once("Deprecated: Instead of strip_spl_white_spaces use replace_spl_white_spaces_with_space")
    return replace_spl_white_spaces_with_space(v)

def replace_spl_white_spaces_with_space(v):
    # check None
    if (v is None):
        return None

    # return
    return str(v).replace("\t", " ").replace("\n", " ").replace("\v", " ").replace("\r", " ")

def resolve_meta_params(xstr, props):
    # check for None
    if (xstr is None):
        return None

    # iterate through properties
    for k in props.keys():
        kstr = "{" + k + "}"
        xstr = xstr.replace(kstr, str(props[k]))

    # return
    return xstr

def replace_template_props(props, xstr):
    # validation
    if (xstr is None or xstr == ""):
        return xstr

    # iterate
    for k in props.keys():
        # generate template key
        template_key = "{" + k + "}"
        if (xstr.find(template_key) != -1):
            xstr = xstr.replace(template_key, str(props[k]))

    # return
    return xstr

def random_shuffle(xs, seed = 0):
    # boundary conditions
    if (xs is None or len(xs) <= 1):
        return xs

    # apply seed
    index = seed % len(xs)

    # create a copy
    xs2 = list([t for t in xs])

    # swap item 0 with item[seed]
    temp = xs2[0]
    xs2[0] = xs2[index]
    xs2[index] = temp
 
    # return
    return xs2

def is_text_content_col(col, text_columns):
    warn_once("is_text_content_col: this api needs more consistency across the board")

    # validation
    if (col is None):
        return False

    # check for text_columns
    if (text_columns is None or len(text_columns) == 0):
        return False

    # use case insensitive matching
    for tcol in text_columns:
        if (col.lower() == tcol.lower()):
            return True

    return False


def resolve_default_parameter(name, value, default_value, msg):
    # check if prefix parameter is None
    if (value is None):
        default_value_display = str(default_value)
        default_value_display = default_value_display if (len(default_value_display) < 30) else default_value_display[0:30] + "..."
        warn_once("{}: {} value is None. Using default value: {}".format(msg, name, default_value_display))
        value = default_value

    # return
    return value

def extend_inherit_message(old_msg, new_msg):
     # check if both msgs are defined
     if (old_msg is not None and new_msg is not None):
         parts1 = list([t.strip() for t in old_msg.split(":")])
         parts2 = list([t.strip() for t in new_msg.split(":")])

         # do a fuzzy match
         if (len(parts2) == 2):
             if (len(parts1) >= 2 and parts1[-2] == parts2[0]):
                 new_msg = parts2[1]
             elif (len(parts1) >= 3 and parts1[-3] == parts2[0]):
                 new_msg = parts2[1]
             elif (len(parts1) >= 4 and parts1[-4] == parts2[0]):
                 new_msg = parts2[1]
             elif (len(parts1) >= 5 and parts1[-5] == parts2[0]):
                 new_msg = parts2[1]

     # return
     return "{}: {}".format(old_msg, new_msg) if (old_msg is not None and len(old_msg) > 0) else "{}".format(new_msg)

def max_dmsg_str(dmsg, max_len = 300):
    if (dmsg is None or len(dmsg) <= max_len):
        return dmsg
    else:
        return "{}... ".format(dmsg[0:max_len - 4])

def is_tsv_file_extension(path):
    # check extensions
    if (path.endswith(".tsv") or path.endswith(".tsv.gz") or path.endswith(".csv") or path.endswith(".csv.gz")):
        return True
    else:
        return False

# TODO: Move this to proper package 
class CombGenerator:
    def __init__(self):
        self.comb1_cache = {}
        self.comb2_cache = {}
        self.comb3_cache = {}
        self.comb4_cache = {}
        self.comb5_cache = {}
        self.comb6_cache = {}
        self.comb7_cache = {}

    def gen_comb1(self, n):
        if (n not in self.comb1_cache.keys()):
            self.comb1_cache[n] = self.__gen_comb1__(n)
        return self.comb1_cache[n]

    def gen_comb2(self, n):
        if (n not in self.comb2_cache.keys()):
            self.comb2_cache[n] = self.__gen_comb2__(n)
        return self.comb2_cache[n]

    def gen_comb3(self, n):
        if (n not in self.comb3_cache.keys()):
            self.comb3_cache[n] = self.__gen_comb3__(n)
        return self.comb3_cache[n]

    def gen_comb4(self, n):
        if (n not in self.comb4_cache.keys()):
            self.comb4_cache[n] = self.__gen_comb4__(n)
        return self.comb4_cache[n]

    def gen_comb5(self, n):
        if (n not in self.comb5_cache.keys()):
            self.comb5_cache[n] = self.__gen_comb5__(n)
        return self.comb5_cache[n]

    def gen_comb6(self, n):
        if (n not in self.comb6_cache.keys()):
            self.comb6_cache[n] = self.__gen_comb6__(n)
        return self.comb6_cache[n]

    def gen_comb7(self, n):
        if (n not in self.comb7_cache.keys()):
            self.comb7_cache[n] = self.__gen_comb7__(n)
        return self.comb7_cache[n]

    def __gen_comb1__(self, n):
       result = []
       for i1 in range(0, n):
           result.append([i1])
       return result
    
    def __gen_comb2__(self, n):
       result = []
       for i2 in range(0, n-1):
           for i1 in range(i2+1, n):
               result.append([i2, i1])
       return result
    
    def __gen_comb3__(self, n):
       result = []
       for i3 in range(0, n-2):
           for i2 in range(i3+1, n-1):
               for i1 in range(i2+1, n):
                   result.append([i3, i2, i1])
       return result
    
    def __gen_comb4__(self, n):
       result = []
       for i4 in range(0, n-3):
           for i3 in range(i4+1, n-2):
               for i2 in range(i3+1, n-1):
                   for i1 in range(i2+1, n):
                       result.append([i4, i3, i2, i1])
       return result
    
    def __gen_comb5__(self, n):
       result = []
       for i5 in range(0, n-4):
           for i4 in range(i5+1, n-3):
               for i3 in range(i4+1, n-2):
                   for i2 in range(i3+1, n-1):
                       for i1 in range(i2+1, n):
                           result.append([i5, i4, i3, i2, i1])
       return result
    
    def __gen_comb6__(self, n):
       result = []
       for i6 in range(0, n-5):
           for i5 in range(i6+1, n-4):
               for i4 in range(i5+1, n-3):
                   for i3 in range(i4+1, n-2):
                       for i2 in range(i3+1, n-1):
                           for i1 in range(i2+1, n):
                               result.append([i6, i5, i4, i3, i2, i1])
       return result
    
    def __gen_comb7__(self, n):
       result = []
       for i7 in range(0, n-6):
           for i6 in range(i7+1, n-5):
               for i5 in range(i6+1, n-4):
                   for i4 in range(i5+1, n-3):
                       for i3 in range(i4+1, n-2):
                           for i2 in range(i3+1, n-1):
                               for i1 in range(i2+1, n):
                                   result.append([i7, i6, i5, i4, i3, i2, i1])
       return result


def split_str_to_arr(x):
    if (x is None or x == ""):
        return []
    else:
        return list(filter(lambda t: t != "", x.split(",")))

    # get string
    result = " ".join(results)

    # return
    return result

def validate_nonnull_params(*args, **kwargs):
    # iterate and throw exception if any of the parameters is None
    for k in kwargs:
        if (kwargs[k] is None):
            raise Exception("Found None value in a mandatory parameter: {}".format(k))

def convert_ipv4_to_hex(ip):
    # validation
    if (ip is None or ip == ""):
        raise Exception("convert_ipv4_to_hex: invalid input: {}".format(ip))

    # split
    parts = ip.split('.')

    # validation
    if (len(parts) != 4):
        raise Exception("convert_ipv4_to_hex: invalid input: {}".format(ip))

    # convert
    hex_parts = hex(int(parts[0])) + hex(int(parts[1])) + hex(int(parts[2])) + hex(int(parts[3]))

    # return
    return hex_parts.replace('0x', '')

def logger_info(logger, msg):
    if (is_info()):
        logger.info(msg)

def logger_warn(logger, msg):
    if (is_warn()):
        logger.warn(msg)



"""library of function to be used in general purpose data processing for TSV class"""

import statistics
import numpy as np
from dateutil import parser
import datetime
from omigo_core import utils

# TODO: mkstr variantgs needs to use *args

def parse_image_file_base_name(x):
    if (len(x) <= 1):
        return x

    index = -1
    if ("\\" in x):
        index = x.rindex("\\")
    elif ("/" in x):
        index = x.rindex("/")
    return str(x[index+1:])

def get_len(vs):
    return str(len(vs))

def get_non_empty_len(vs):
    vs = list(filter(lambda t: len(t.strip()) > 0, vs))
    return str(len(vs))

def uniq_len(vs):
    vs2 = set()
    for t in vs:
        for k in str(t).split(","):
            if (len(k.strip()) > 0):
                vs2.add(str(k))
    return str(len(vs2))

def uniq_mkstr(vs):
    vs2 = set()
    for t in vs:
        for k in str(t).split(","):
            if (len(k.strip()) > 0):
                vs2.add(str(k))
    return ",".join(sorted([str(x) for x in vs2]))

def split_merge_uniq_mkstr(vs):
    vs2 = []
    for v in vs:
        vs2 = vs2 + v.split(",")
    vs2 = list(set(filter(lambda t: len(t.strip()) > 0, vs2)))
    return ",".join(sorted([str(x) for x in vs2]))

def mean(vs):
    vs = list([float(v) for v in vs])
    return statistics.mean(vs)

def std_dev(vs):
    vs = list([float(v) for v in vs])

    # adding check for atleast 2 elements before calling the statistics package
    if (len(vs) < 2):
        return 0

    # return
    return statistics.stdev(vs)

def mkstr(vs):
    vs2 = list(filter(lambda t: len(t.strip()) > 0, [str(x) for x in vs]))
    return ",".join(vs2)

def sorted_mkstr(vs):
    vs2 = sorted(list(filter(lambda t: len(t.strip()) > 0, [str(x) for x in vs])))
    return ",".join(vs2)

def mkstr4f(vs):
    vs2 = list(["{:4f}".format(float(x)) for x in vs])
    return ",".join(vs2)

def minint(vs):
    if (vs is None or len(vs) == 0):
        raise Exception("minint: empty vs")

    min_value = str(vs[0])
    for v in vs[1:]:
        if (int(float(v)) < int(float(min_value))):
            min_value = str(v)

    return str(min_value)

def maxint(vs):
    if (vs is None or len(vs) == 0):
        raise Exception("maxint: empty vs")

    max_value = str(vs[0])
    for v in vs[1:]:
        if (int(float(v)) > int(float(max_value))):
            max_value = str(v)

    return str(max_value)

def minfloat(vs):
    if (vs is None or len(vs) == 0):
        raise Exception("minfloat: empty vs")

    min_value = str(vs[0])
    for v in vs[1:]:
        if (float(v) < float(min_value)):
            min_value = str(v)

    return str(min_value)

def maxfloat(vs):
    if (vs is None or len(vs) == 0):
        raise Exception("maxfloat: empty vs")

    max_value = str(vs[0])
    for v in vs[1:]:
        if (float(v) > float(max_value)):
            max_value = str(v)

    return str(max_value)

def minstr(vs):
    if (vs is None or len(vs) == 0):
        raise Exception("minstr: empty vs")

    min_value = str(vs[0])
    for v in vs[1:]:
        if (str(v) < min_value):
            min_value = str(v)

    return str(min_value)

def maxstr(vs):
    if (vs is None or len(vs) == 0):
        raise Exception("maxstr: empty vs")

    max_value = str(vs[0])
    for v in vs[1:]:
        if (str(v) > max_value):
            max_value = str(v)

    return str(max_value)

def minint_failsafe(vs):
    if (vs is None or len(vs) == 0):
        return "" 
    else:
        vs = list(filter(lambda t: t != "", vs))
        if (len(vs) == 0):
            return "" 
        else:
            return minint(vs) 

def maxint_failsafe(vs):
    if (vs is None or len(vs) == 0):
        return "" 
    else:
        vs = list(filter(lambda t: t != "", vs))
        if (len(vs) == 0):
            return "" 
        else:
            return maxint(vs) 

def minstr_failsafe(vs):
    if (vs is None or len(vs) == 0):
        return "" 
    else:
        vs = list(filter(lambda t: t != "", vs))
        if (len(vs) == 0):
            return "" 
        else:
            return minstr(vs) 
            
def maxstr_failsafe(vs):
    if (vs is None or len(vs) == 0):
        return "" 
    else:
        vs = list(filter(lambda t: t != "", vs))
        if (len(vs) == 0):
            return "" 
        else:
            return maxstr(vs)

def sumint(vs):
    if (len(vs) == 0):
        return 0
    else:
        return sum([int(float(t)) for t in vs])

def sumfloat(vs):
    if (len(vs) == 0):
        return 0.0
    else:
        return sum([float(t) for t in vs])

# TODO: The semantics are not clear
def uniq_count(vs):
    vs2 = list(filter(lambda t: t.strip() != "", vs))
    return len(set(vs2))

def merge_uniq(vs):
    result = []
    for v in vs:
        for v2 in v.split(","):
            result.append(v2)

    # return
    return ",".join(sorted(list(set(result))))

def select_first(x, y):
    return x

def select_max_int(x, y):
    return int(max(int(x), int(y)))

def str_arr_to_float(xs):
    return [float(x) for x in xs]

def quantile(xs, start = 0, end = 1, by = 0.25, precision = 4):
    if (start > end):
        raise Exception("Start: {} > End: {}".format(start, end))

    qarr = []
    cur = start
    while (cur < end):
        qarr.append(cur)
        cur = cur + by

    format_str = "{:." + str(precision) + "f}"
    quan = np.quantile(str_arr_to_float(xs), qarr)
    return ",".join(format_str.format(x) for x in quan)

def quantile4(xs):
    return quantile(xs)

def quantile10(xs):
    return quantile(xs, by=1/10)

def quantile40(xs):
    return quantile(xs, by=1/40)

def max_str(xs):
    utils.warn_once("max_str is deprecated. Use maxstr")
    xs = sorted(xs)
    return xs[-1]

def min_str(xs):
    utils.warn_once("min_str is deprecated. Use minstr")
    xs = sorted(xs)
    return xs[0]

def to2digit(x):
    return "{:.2f}".format(float(x))

def to4digit(x):
    return "{:.4f}".format(float(x))

def to6digit(x):
    return "{:.6f}".format(float(x))

def convert_prob_to_binary(x, split=0.5):
    if (x >= split):
        return 1
    else:
        return 0

def get_str_map_with_keys(mp, keys, fail_on_missing = True):
    mp2 = {}
    for k in keys:
        if (k in mp.keys()):
            mp2[k] = str(mp[k])
        else:
            if (fail_on_missing):
                raise Exception("Missing key: {}, avaialble keys: {}".format(k, str(mp.keys())))

    return mp2

def get_str_map_without_keys(mp, excluded_keys):
    mp2 = {}
    for k in mp.keys():
        if (k not in excluded_keys):
            mp2[k] = str(mp[k])

    return mp2

def datetime_to_utctimestamp_millis(x):
    # convert this to string first for failsafe
    x = str(x)

    # 1681202675933
    if (len(str(x)) == 13 and str(x).isnumeric() == True):
        # this looks like numeric timestamp in millis
        return int(x)

    # 1681202675.933
    if (len(str(x)) == 14 and str(x).find(".") == 10 and str(x).isnumeric() == True):
        # this looks like numeric timestamp in millis
        return int(x)

    # 2023-04-11T08:44:35.933Z
    if (len(x) == 24 and x[19] == "." and x.endswith("Z")):
        return int(float(parser.parse(x).timestamp() * 1000))

    # 2023-04-15T15:05:16.175000Z
    if (len(x) == 27 and x[19] == "." and x.endswith("Z")):
        return int(float(parser.parse(x).timestamp() * 1000))

    # 2023-04-11T08:44:35.933+00:00
    if (len(x) == 29 and x[-6] == "+"):
        return int(float(parser.parse(x).timestamp() * 1000))

    # 2023-04-18T18:47:45 or 2023-04-18 18:47:45
    if (len(x) == 19 and (x[10] == "T" or x[10] == " ")):
        return int(float(parser.parse(x + "+00:00").timestamp() * 1000))

    # this seems to be a timestamp with second precision.
    return int(datetime_to_utctimestamp_sec(x) * 1000)

# TODO. better naming
def datetime_to_utctimestamp(x):
    # use datetime_to_utctimestamp_sec
    utils.warn_once("datetime_to_utctimestamp: Deprecated. Use datetime_to_utctimestamp_sec instead")
    return datetime_to_utctimestamp_sec(x)

def datetime_to_utctimestamp_sec(x):
    # convert this to string first for failsafe
    x = str(x)

    # 2022-05-20T05:00:00+00:00
    if (x.endswith("UTC") or x.endswith("GMT") or x.endswith("Z") or x.endswith("+00:00")):
        return int(parser.parse(x).timestamp())
    elif (len(x) == 10 and x.find("-") != -1):
        # 2021-11-01
        x = x + "T00:00:00Z"
        return int(parser.parse(x).timestamp())
    elif (len(x) == 19):
        # 2021-11-01T00:00:00
        x = x + "Z"
        if (x[10] == " "):
            x = x.replace(" ", "T")
        return int(parser.parse(x).timestamp())
    elif (len(x) == 23 and x[19] == "."):
        # 2021-11-01T00:00:00.000
        x = x + "Z"
        return int(parser.parse(x).timestamp())
    elif (len(x) == 26):
        # 2021-11-01T00:00:00.000000
        x = x + "Z"
        return int(parser.parse(x).timestamp())
    elif (len(x) == 27 and x[19] == "." and x.endswith("Z")):
        # 2023-04-15T15:05:16.175000Z
        return int(parser.parse(x).timestamp())
    elif (len(x) == 29 and x[-6] == "+"):
        # 2023-04-11T08:44:35.933+00:00
        return int(parser.parse(x).timestamp())
    elif (len(x) == 10 and str(x).isnumeric() == True):
        # this looks like a numeric timestamp
        return int(x)
    elif (len(x) == 13 and str(x).isnumeric() == True):
        # this looks like numeric timestamp in millis
        return int(int(x) / 1000)
    elif (len(x) == 19 and (x[10] == "T" or x[10] == " ")):
        # 2023-04-18T18:47:45 or 2023-04-18 18:47:45
        return int(float(parser.parse(x + "+00:00").timestamp() * 1000))
    else:
        raise Exception("Unknown date format. Problem with UTC: '{}'".format(x))

# TODO: Converts seconds format only. Even the original time in milliseconds will return seconds format
def utctimestamp_to_datetime(x):
    # use the string form
    x = str(x)
    if (len(x) == 10 and x.isnumeric() == True):
        return datetime.datetime.utcfromtimestamp(int(x)).replace(tzinfo = datetime.timezone.utc)
    elif (len(x) == 13 and x.isnumeric() == True):
        return datetime.datetime.utcfromtimestamp(int(x)/1000).replace(tzinfo = datetime.timezone.utc)
    elif (len(x) > 10 and x.find(".") == 10 and utils.is_float(x)): 
        return datetime.datetime.utcfromtimestamp(float(x)).replace(tzinfo = datetime.timezone.utc)
    else:
        raise Exception("Unknown timestamp format: {}".format(x))

# TODO: Converts seconds format only
def utctimestamp_millis_to_datetime(x):
    return utctimestamp_to_datetime(x)

# TODO: Converts seconds format only
# Its utc so removed the last timezone
def utctimestamp_to_datetime_str(x):
    return utctimestamp_to_datetime(x).isoformat()[0:19]

# Its utc so removed the last timezone. TODO: Keep the UTC or Z
def utctimestamp_millis_to_datetime_str(x):
    result = utctimestamp_to_datetime(x).isoformat()
    if (result.endswith("UTC")):
        return result[0:23]
    else:
        return result

def datetime_to_timestamp(x):
    raise Exception("Please use datetime_to_utctimestamp")

def get_utctimestamp_sec():
    return int(datetime.datetime.now(datetime.timezone.utc).timestamp())

def get_utctimestamp_millis():
    return int(datetime.datetime.now(datetime.timezone.utc).timestamp() * 1000)

def datestr_to_datetime(x):
    return utctimestamp_to_datetime(datetime_to_utctimestamp_sec(x))

def select_first_non_empty(*args, **kwargs):
    # variable name
    xs = args

    # boundary condition
    if (xs is None or len(xs) == 0):
        return ""

    # check if it is list or tuple
    if (isinstance(xs[0], list)):
        xs = xs[0]

    # boundary conditions
    for x in xs:
        if (x is not None and x != ""):
            return str(x)

    # return default
    return ""

# def if_else_non_empty_str(x, v1, v2):
def if_else_non_empty_str(*args):
    # variable name
    xs = args

    # boundary condition
    if (xs is None or len(xs) == 0):
        raise Exception("if_else_non_empty_str: invalid input") 

    # check if it is list or tuple
    if (isinstance(xs[0], list)):
        xs = xs[0]

    # validation
    if (len(xs) != 3):
        raise Exception("if_else_non_empty_str: invalid input: {}".format(xs))

    # assign variables
    x = xs[0]
    v1 = xs[1]
    v2 = xs[2]

    if (x is not None and str(x) != ""):
        return str(v1)
    else:
        return str(v2)

# def if_else_str(xinput, xval, v1, v2):
def if_else_str(*args):
    # variable name
    xs = args

    # boundary condition
    if (xs is None or len(xs) == 0):
        raise Exception("if_else_non_empty_str: invalid input") 

    # check if it is list or tuple
    if (isinstance(xs[0], list)):
        xs = xs[0]

    # validation
    if (len(xs) != 4):
        raise Exception("if_else_str: invalid input: {}".format(xs))

    # assign variables
    xinput = xs[0]
    xval = xs[1]
    v1 = xs[2]
    v2 = xs[3]

    # apply function
    if (xinput is not None and str(xinput) == str(xval)):
        return str(v1)
    else:
        return str(v2)

# def if_else_int(xinput, xval, v1, v2):
def if_else_int(*args):
    # variable name
    xs = args

    # boundary condition
    if (xs is None or len(xs) == 0):
        raise Exception("if_else_int: invalid input") 

    # check if it is list or tuple
    if (isinstance(xs[0], list)):
        xs = xs[0]

    # validation
    if (len(xs) != 4):
        raise Exception("if_else_int: invalid input: {}".format(xs))

    # assign variables
    xinput = xs[0]
    xval = xs[1]
    v1 = xs[2]
    v2 = xs[3]

    # apply function
    if (xinput is not None and int(xinput) == int(xval)):
        return v1
    else:
        return v2

# def if_else_non_zero_int(x, v1, v2):
def if_else_non_zero_int(*args):
    # variable name
    xs = args

    # boundary condition
    if (xs is None or len(xs) == 0):
        raise Exception("if_else_non_zero_int: invalid input") 

    # check if it is list or tuple
    if (isinstance(xs[0], list)):
        xs = xs[0]

    # validation
    if (len(xs) != 3):
        raise Exception("if_else_int: invalid input: {}".format(xs))

    # assign variables
    x = xs[0]
    v1 = xs[1]
    v2 = xs[2]

    # apply function
    if (x is None or int(x) != 0):
        return v1
    else:
        return v2

# TODO: this is bad implementation. The Win32 timestamp format needs proper handling
def win32_timestamp_to_utctimestamp(x):
    return int(str(x)[0:-8]) + 339576461

def get_time_diffs(vs):
    # sort the input
    vs = sorted(list([datetime_to_utctimestamp_sec(t) for t in vs]))

    # boundary condition
    if (len(vs) <= 1):
        return ""

    # get parirs
    pairs = list(zip(vs[0:len(vs)-1], vs[1:]))

    # result
    result = []
    for pair in pairs:
        v1, v2 = pair
        diff = int(v2 - v1)
        if (diff < 60):
            result.append("{}s".format(diff))
        elif (diff < 60 * 60):
            result.append("{}m".format(int(diff / 60)))
        elif (diff < 24 * 60 * 60):
            result.append("{}h".format(int(diff / (60 * 60))))
        elif (diff < 30 * 24 * 60 * 60):
            result.append("{}d".format(int(diff / (24 * 60 * 60))))
        else:
            result.append("{}:s".format(diff))

    # return
    return ",".join(result)

def simple_map_to_url_encoded_col_names(cols, url_encoded_cols = None):
    # create result
    results = []
    if (url_encoded_cols is not None):
        # iterate
        for c in cols:
            if (c in url_encoded_cols):
                results.append("{}:url_encoded".format(c))
            else:
                results.append(c)
    else:
        results = cols

    # return
    return results

def map_to_url_encoded_col_names(cols, prefix = None, url_encoded_cols = None):
    results = []

    # iterate
    for c in cols:
        col = c
        # check if there is prefix 
        if (col.find(":") != -1):
            col = col.split(":")[-1]

        # assign
        result = c

        # handle url_encoded suffix
        if (url_encoded_cols is not None and col in url_encoded_cols):
            result = "{}:url_encoded".format(c)

        # handle prefix
        if (prefix is not None):
            result = "{}:{}".format(prefix, result)

        # append
        results.append(result)

    # return
    return results

def get_display_relative_time_str(v):
    # convert to int
    v = int(v)

    # compute units
    days = v // 86400
    hours = (v - (days * 86400)) // 3600
    minutes = (v - (days * 86400 + hours * 3600)) // 60
    seconds = v - (days * 86400 + hours * 3600 + minutes * 60)
            
    # compute display string
    results = []
    count = 0
    max_display_values = 2
    if (days > 0 and count < max_display_values):
        results.append("{}d".format(days))
        count = count + 1
    if (hours > 0 and count < max_display_values):
        results.append("{}h".format(hours))
        count = count + 1
    if (minutes > 0 and count < max_display_values):
        results.append("{}m".format(minutes))
        count = count + 1
    if (seconds > 0 and count < max_display_values):
        results.append("{}s".format(seconds))
        count = count + 1

    # get string
    result = " ".join(results)

    # return
    return result

